-- This file contains functions for enabling/disabling cheats.
if not cheats then cheats = {} end

-- Returns whether Creative Mode has been enabled. If not, a message will be printed for the given player.
function cheats.check_creative_mode_has_enabled(player)
	-- If creative mode has not been enabled, notify the player in charge and do nothing.
	if not global.creative_mode.enabled then
		if player then
			player.print({"message.creative-mode_not-yet-enabled"})
		end
		return false
	end
	return true
end

-- Returns whether Creative Mode has not been enabled. If so, a message will be printed for the given player.
function cheats.check_creative_mode_has_not_enabled(player)
	if global.creative_mode.enabled then
		if player then
			player.print({"message.creative-mode_already-enabled"})
		end
		return false
	end
	return true
end

-- Enables or disables Creative Mode for all players.
-- @param host	The player who enables or disables Creative Mode.
function cheats.enable_or_disable_creative_mode(enable, is_permanent, auto_apply_to_all_cheats, host)
	if enable then
		-- If creative mode has been enabled, notify the player in charge and do nothing.
		if not cheats.check_creative_mode_has_not_enabled(host) then
			return
		end
		
		-- If creative mode has already been permanently disabled, notify the player in charge and do nothing.
		if global.creative_mode.permanently_disabled then
			if host then
				host.print({"message.creative-mode_enable-failed"})
			end
			return
		end
	else
		-- If creative mode has not been enabled, notify the player in charge and do nothing.
		if not cheats.check_creative_mode_has_enabled(host) then
			return
		end
	end
	
	-- Enable/Disable the recipes.
	cheats.enable_or_disable_creative_recipes(enable)
	
	-- Also unhide/hide the items so they can be selected as filter.
	-- Unfortunately, LuaItemPrototype doesn't have flags property. This cannot be done at this moment.
--		for _, item in pairs(creative_mode_defines.names.items) do
--			if game.item_prototypes[item] ~= nil then
--				if enable then
--					table.remove(game.item_prototypes[item].flags, 2) -- The hidden flag is the second one.
--				else
--					table.insert(game.item_prototypes[item].flags, "hidden")
--				end
--			end
--		end
	
	local player_name
	if host.name == nil then
		player_name = {"entity-name.player"}
	else
		player_name = host.name
	end
		
	if enable then
		-- Notify the players.
		util.print_to_all_players({"message.creative-mode_enable-succeeded", player_name})
		
		-- Also raise the custom event.
		remote_interface.raise_on_enabled_event()
	else
		-- Notify the players.
		if is_permanent then
			util.print_to_all_players({"message.creative-mode_permanently-disable-succeeded", player_name})
		else
			util.print_to_all_players({"message.creative-mode_disable-succeeded", player_name})
		end
		
		-- Also raise the custom event.
		remote_interface.raise_on_disabled_event()
	end
	
	-- Save as creative mode has been enabled/disabled.
	global.creative_mode.enabled = enable
	
	-- If it is permanently disabled, save the state so it cannot be enabled later.
	if not enable and is_permanent then
		global.creative_mode.permanently_disabled = true
	end
	
	-- Apply to all cheats if needed
	if auto_apply_to_all_cheats then
		cheats.enable_or_disable_cheat_mode(enable)
		cheats.enable_or_disable_loaders(enable)
		cheats.research_or_reset_all_technologies(enable)
		cheats.enable_or_disable_always_day(enable)
		cheats.increase_or_reset_character_reach(enable)
		cheats.increase_or_reset_character_mining_speed(enable)
		cheats.increase_or_reset_character_running_speed(enable)
		cheats.enable_or_disable_instant_blueprint(enable)
		cheats.enable_or_disable_instant_deconstruction(enable)
	else
		-- Show popup for enabling/disabling more features.
		if enable then
			gui.show_main_popup(host, gui.main_popup_content_type.enable_more_cheats)
		else
			gui.show_main_popup(host, gui.main_popup_content_type.disable_more_cheats)
		end
	end
end

-- Enables or disables recipes for our creative items.
function cheats.enable_or_disable_creative_recipes(enable)
	for _, force in pairs(game.forces) do
		-- Item recipes.
		for _, recipe in pairs(creative_mode_defines.names.recipes) do
			if force.recipes[recipe] ~= nil then
				force.recipes[recipe].enabled = enable
			end
		end
		-- Free fluid recipes. Get the recipe names by looping through the fluid prototypes in game.
		for _, fluid in pairs(game.fluid_prototypes) do
			force.recipes[creative_mode_defines.names.free_fluid_recipe_prefix .. fluid.name].enabled = enable
		end
	end
end

-- Enables or disables cheat mode for all players.
function cheats.enable_or_disable_cheat_mode(enable)
	for _, player in pairs(game.players) do
		player.cheat_mode = enable
		if enable then
			player.print({"message.creative-mode_cheat-mode-enabled"})
		else
			player.print({"message.creative-mode_cheat-mode-disabled"})
		end
	end
end

-- Enables or disables the recipes of loaders.
function cheats.enable_or_disable_loaders(enable)
	for _, force in pairs(game.forces) do
		for _, recipe in pairs(creative_mode_defines.names.loader_recipes) do
			if force.recipes[recipe] ~= nil then
				force.recipes[recipe].enabled = enable
			end
		end
	end
	if enable then
		util.print_to_all_players({"message.creative-mode_loaders-enabled"})
	else
		util.print_to_all_players({"message.creative-mode_loaders-disabled"})
	end
end

-- Researches or resets all technologies for all forces.
function cheats.research_or_reset_all_technologies(research)
	for _, force in pairs(game.forces) do
		if research then
			force.research_all_technologies()
		else
			force.reset()
		end
	end
	if research then
		util.print_to_all_players({"message.creative-mode_all-technologies-researched"})
	else
		util.print_to_all_players({"message.creative-mode_all-technologies-reset"})
	end
end

-- Makes all surfaces always day or restores day night cycle on all surfaces.
function cheats.enable_or_disable_always_day(enable)
	for _, surface in pairs(game.surfaces) do
		surface.always_day = enable
	end
	if enable then
		util.print_to_all_players({"message.creative-mode_made-all-surfaces-always-day"})
	else
		util.print_to_all_players({"message.creative-mode_brought-back-night-to-surfaces"})
	end
end

-- Increases or resets the reach distances of all players' characters.
function cheats.increase_or_reset_character_reach(increase)
	for _, player in pairs(game.players) do
		if player.character then
			if increase then
				player.character_build_distance_bonus = 10000000
				player.character_item_drop_distance_bonus = 10000000
				player.character_reach_distance_bonus = 10000000
				player.character_resource_reach_distance_bonus = 10000000
				player.print({"message.creative-mode_character-reach-increased"})
			else
				player.character_build_distance_bonus = 0
				player.character_item_drop_distance_bonus = 0
				player.character_reach_distance_bonus = 0
				player.character_resource_reach_distance_bonus = 0
				player.print({"message.creative-mode_character-reach-reset"})
			end
		end
	end
end

-- Increases or resets the mining speed of all players' characters.
function cheats.increase_or_reset_character_mining_speed(increase)
	for _, player in pairs(game.players) do
		if player.character then
			if increase then
				player.character_mining_speed_modifier = 100
				player.print({"message.creative-mode_character-mining-speed-increased"})
			else
				player.character_mining_speed_modifier = 0
				player.print({"message.creative-mode_character-mining-speed-reset"})
			end
		end
	end
end

-- Increases the running speed of all players' characters.
function cheats.increase_or_reset_character_running_speed(increase)
	for _, player in pairs(game.players) do
		if player.character then
			if increase then
				player.character_running_speed_modifier = 2
				player.print({"message.creative-mode_character-running-speed-increased"})
			else
				player.character_running_speed_modifier = 0
				player.print({"message.creative-mode_character-running-speed-reset"})
			end
		end
	end
end

-- Enables or disables instant blueprint.
function cheats.enable_or_disable_instant_blueprint(enable)
	global.creative_mode.cheats.instant_blueprint = enable
	if enable then
		util.print_to_all_players({"message.creative-mode_instant-blueprint-enabled"})
	else
		util.print_to_all_players({"message.creative-mode_instant-blueprint-disabled"})
	end
end

-- Enables or disables instant deconstruction.
function cheats.enable_or_disable_instant_deconstruction(enable)
	global.creative_mode.cheats.instant_deconstruction = enable
	if enable then
		util.print_to_all_players({"message.creative-mode_instant-deconstruction-enabled"})
	else
		util.print_to_all_players({"message.creative-mode_instant-deconstruction-disabled"})
	end
end