-- This file contains variables or functions that are related to the Creative Chest family in this mod.
if not creative_chest_util then creative_chest_util = {} end

-- All possible inventory display mode for creative chests.
creative_chest_util.inventory_display_modes =
{
	original_mode = 1,
	compact_mode = 2,
}

-- Some of the variables are under the namespace of Creative Provider Chest because it was implemented before Creative Chest.

-- Updates the data used by the Creative Chest family according to the in-game item list in global.
function creative_chest_util.update_item_list_data()
	-- There may be more items than the usable inventory size of each creative chest. We need to find out how many chests are needed for one cycle first.
	-- It is also the total number of item groups.
	global.creative_mode.creative_provider_chest_num_in_cycle = math.ceil(#global.item_list / creative_mode_config.creative_provider_chest.inventory_size)
	-- Chests in the same cycle are divided into groups. Let's say we have 251 items, so the first group will contain the first 250 items, and the last group will contain the last item.
	global.creative_mode.creative_provider_chest_next_place_group = 1
	global.creative_mode.creative_chest_next_place_group = 1
	-- For performance reason, we don't update all groups in the same tick.
	global.creative_mode.creative_provider_chest_next_update_group = 1
	global.creative_mode.creative_chest_next_update_group = 1
	-- And we don't update all chests in the same group.
	global.creative_mode.creative_provider_chest_next_update_group_subindex = 1
	global.creative_mode.creative_chest_next_update_group_subindex = 1
	
	-- Store the settings that can affect the number of chests per item cycle when we have calculated the above maths, so that we know we will need to calculate again if change is detected.
	global.creative_mode.last_creative_provider_chest_size = creative_mode_config.creative_provider_chest.inventory_size
	global.creative_mode.last_creative_provider_chest_contain_hidden_items = global.creative_mode.config.creative_provider_chest.contain_hidden_items
end

-- Validates the data used by the Creative Chest family. If needed, updates the item list and also the data.
-- Returns whether the data is updated.
local function validate_or_update_item_list_data()
	if (global.creative_mode.last_creative_provider_chest_size == nil or global.creative_mode.last_creative_provider_chest_size ~= creative_mode_config.creative_provider_chest.inventory_size or
		global.creative_mode.last_creative_provider_chest_contain_hidden_items == nil or global.creative_mode.last_creative_provider_chest_contain_hidden_items ~= global.creative_mode.config.creative_provider_chest.contain_hidden_items) then
		global_util.renew_item_list()
		return true
	end
	return false
end

-- Processes the tables related to the Creative Chest family in global.
function creative_chest_util.tick()
	-- Check whether the user has changed the settings that can affect the number of chests per item cycle. If so, update the item list and regroup the items.
	if game.tick >= 1 then 
		if validate_or_update_item_list_data() then
			util.print_to_all_players({"message.creative-mode_creative-provider-chest-item-group-updated", global.creative_mode.creative_provider_chest_num_in_cycle})
		end
	end
end

-- Returns the start and end item indexes among the item list for the group of given group number.
function creative_chest_util.get_start_end_item_index_for_group(group_number)
	local start_item_index = (group_number - 1) * creative_mode_config.creative_provider_chest.inventory_size + 1
	local end_item_index = math.min(group_number * creative_mode_config.creative_provider_chest.inventory_size, #global.item_list)
	return start_item_index, end_item_index
end

-- Refills the inventories of some Creative Chests according to the given chest groups, group index to be updated, and the subindex of chest to be updated.
-- Returns the updated group index and subindex.
function creative_chest_util.refill_chests(chest_data_groups, next_update_group, next_update_group_subindex)
	-- Looping through the huge item list can be a heavy task, therefore we divided the task by groups and chests.
	-- Only the group of the current group number has to be updated.
	if chest_data_groups[next_update_group] then
		local start_item_index, end_item_index = creative_chest_util.get_start_end_item_index_for_group(next_update_group)
		-- Not even that, we also update only one chest in each tick.
		local data = chest_data_groups[next_update_group][next_update_group_subindex]
		if data then
			local chest = data.entity
			if chest.valid then
				if not chest.to_be_deconstructed(chest.force) then
					local inventory = chest.get_output_inventory()
					local slot = 1
					local displayed_slot = 1
					local display_mode = data.inventory_display_mode
					local inventory_size = #inventory
					local filtered_slots = data.filtered_slots
					local filtered_slots_count = #filtered_slots
					local next_filtered_slot_to_check = 1
					if filtered_slots_count <= 0 then
						next_filtered_slot_to_check = 0
					end
					-- Refill its inventory with items.
					for i = start_item_index, end_item_index, 1 do
						-- Fill the slot only if it is not filtered out.
						if next_filtered_slot_to_check <= 0 or filtered_slots[next_filtered_slot_to_check] ~= slot then
							local item = global.item_list[i]
							inventory[displayed_slot].set_stack{name = item.name, count = item.stack_size}
							displayed_slot = displayed_slot + 1
						else
							-- Clear the slot. Only Original Display Mode will create holes in between.
							if display_mode == creative_chest_util.inventory_display_modes.original_mode then
								inventory[displayed_slot].clear()
								displayed_slot = displayed_slot + 1
							end
							-- Check for the next index of filtered slot if possible.
							if filtered_slots_count >= next_filtered_slot_to_check + 1 then
								next_filtered_slot_to_check = next_filtered_slot_to_check + 1
							else
								next_filtered_slot_to_check = 0
							end
						end
						-- Next slot.
						slot = slot + 1
						-- Though it is unlikely to happen... (caused by mod version change?) Break the loop if slot exceeds.
						if slot > inventory_size then
							break
						end
					end
					-- Clear the remaining slots.
					for i = displayed_slot, inventory_size, 1 do
						inventory[i].clear()
					end
				end
			else
				table.remove(chest_data_groups[next_update_group], next_update_group_subindex)
			end
		else
			table.remove(chest_data_groups[next_update_group], next_update_group_subindex)
		end
	end
	-- Prepare the subindex for the next chest in the same group.
	if chest_data_groups[next_update_group] and next_update_group_subindex < #chest_data_groups[next_update_group] then
		next_update_group_subindex = next_update_group_subindex + 1
	else
		-- Next group.
		next_update_group_subindex = 1
		if next_update_group < global.creative_mode.creative_provider_chest_num_in_cycle then
			next_update_group = next_update_group + 1
		else
			next_update_group = 1
		end
	end
	
	return next_update_group, next_update_group_subindex
end


-- Returns the responsible creative chest data groups for the given entity.
function creative_chest_util.get_creative_chest_data_groups(entity)
	if entity.name == creative_mode_defines.names.entities.creative_chest then
		return global.creative_mode.creative_chest_data_groups
	elseif entity.name == creative_mode_defines.names.entities.creative_provider_chest then
		return global.creative_mode.creative_provider_chest_data_groups
	end
	return nil
end

-- Returns the data and group number of the given Creative Chest. Also returns the index of the chest in that group as the second parameter.
-- If the given entity is not found in any group (maybe caused by error), 0, 0 will be returned.
function creative_chest_util.get_creative_chest_data_group_number(entity, chest_data_groups)
	if chest_data_groups then
		for group_index = 1, #chest_data_groups, 1 do
			local group = chest_data_groups[group_index]
			if group then
				for chest_index_in_group, chest_data_in_group in ipairs(chest_data_groups[group_index]) do
					if chest_data_in_group.entity == entity then
						return chest_data_in_group, group_index, chest_index_in_group
					end
				end
			end
		end
	end
	return nil, 0, 0
end

-- Moves the destination chest to the group of the source chest when copy-paste is done on Creative Provider Chest.
-- Also copies the filtered slots from the source chest to the destination chest.
function creative_chest_util.on_creative_chest_pasted(source_chest, destination_chest)
	local source_groups = creative_chest_util.get_creative_chest_data_groups(source_chest)
	local destination_groups = creative_chest_util.get_creative_chest_data_groups(destination_chest)

	local source_chest_data, source_group_number = creative_chest_util.get_creative_chest_data_group_number(source_chest, source_groups)
	if source_chest_data == nil or source_group_number == 0 then
		return
	end
	local destination_chest_data, destination_group_number, destination_index_in_group = creative_chest_util.get_creative_chest_data_group_number(destination_chest, destination_groups)
	if destination_chest_data == nil or destination_group_number == 0 then
		return
	end
	-- Copy the table of filtered slots from the source chest to the destination chest.
	destination_chest_data.filtered_slots = table.deepcopy(source_chest_data.filtered_slots)
	-- Copy inventory display mode.
	destination_chest_data.inventory_display_mode = source_chest_data.inventory_display_mode
	-- Move the destination chest from its original group to the group with the same group number of the source chest.
	if source_group_number ~= destination_group_number then
		table.remove(destination_groups[destination_group_number], destination_index_in_group)
		-- But it should still be inside the original groups table. (One table for non-logistic, one for logistic)
		if not destination_groups[source_group_number] then
			destination_groups[source_group_number] = {}
		end
		table.insert(destination_groups[source_group_number], destination_chest_data)
	end
end

-- Changes the group number of the given creative chest data from the old one to new one.
-- Returns whether it is successfully changed.
function creative_chest_util.change_creative_chest_group_number(chest_data, chest_groups, old_group_number, old_index_in_group, new_group_number)
	if old_group_number ~= new_group_number then
		table.remove(chest_groups[old_group_number], old_index_in_group)
		if not chest_groups[new_group_number] then
			chest_groups[new_group_number] = {}
		end
		table.insert(chest_groups[new_group_number], chest_data)
		return true
	end
	return false
end