-- This file manages event-related variables or functions that are specific for this mod.
-- But the GUI related events are in gui.lua instead.
if not events then events = {} end

-- Initializes values that are shared across files. Most likely clamping is done here.
local function init_shared_values()
	-- Makes sure the inventory size is valid.
	creative_mode_config.creative_provider_chest.inventory_size = util.clamp(creative_mode_config.creative_provider_chest.inventory_size, 1, 254)
end

-- Callback of the on_init event, which is invoked when a new game is started.
function events.on_init()
	init_shared_values()
	
	-- Initializes meta-table.
	global_util.initialize_or_update_global()
	
	-- Generate event IDs if they are not generated.
	remote_interface.get_or_load_on_enabled_event_id()
	remote_interface.get_or_load_on_disabled_event_id()
	
	-- Initializes the item list.
	global_util.initialize_item_list_if_not_exist()
end

-- Callback of the on_load event, which is invoked when an existing game is loaded.
-- The global table should not be touched here.
function events.on_load(event)
	-- Generate event IDs if they are not generated.
	remote_interface.get_or_load_on_enabled_event_id()
	remote_interface.get_or_load_on_disabled_event_id()
end

-- Callback of the on_configuration_changed event, which is invoked when the game version or any active mods changed.
function events.on_configuration_changed(data)
	init_shared_values()

	-- Game version changed?
	local isGameChanged = data.old_version ~= data.new_version
	-- Any mod changed?
	local isAnyModChanged = data.mod_changes ~= nil
	-- Is this mod changed?
	local isOurModChanged = false
	if isAnyModChanged then
		isOurModChanged = data.mod_changes[creative_mode_defines.mod_id] ~= nil
	end
	-- Is this mod just installed for an existing game?
	local isOurModNew = false
	if isOurModChanged then
		isOurModNew = data.mod_changes[creative_mode_defines.mod_id].old_version == nil and data.mod_changes[creative_mode_defines.mod_id].new_version ~= nil
	end

	-- If our mod has changed (whether version is different or the mod is new for existing game), new meta-tables may be needed. Initializes them if needed.
	if isOurModChanged then
		global_util.initialize_or_update_global()
	end
	
	-- For game version changed or any mod changed, item list needs to be updated.
	if global.item_list == nil or isGameChanged or isAnyModChanged then
		global_util.renew_item_list()
	end
	
	-- If creative mode is already enabled, we will also need to make sure all of our recipes are enabled in case there is new recipes for new items.
	if isGameChanged or isAnyModChanged then
		if global.creative_mode.enabled then
			cheats.enable_or_disable_creative_recipes(true)
		end
	end
end

-- Callback of the on_built_entity and on_robot_built_entity events, which are called when an entity is built by either player or robot.
function events.on_built_entity(event)
	-- Thanks Nexela for the inspiration of following snippet.
	-- Instant blueprint, raise created entity event with new entity.
	if global.creative_mode.cheats.instant_blueprint then
		local is_entity_ghost = event.created_entity.name == "entity-ghost"
		local is_tile_ghost = event.created_entity.name == "tile-ghost"
		if is_entity_ghost or is_tile_ghost then
			-- Make sure player_index is provided. It is not likely to be missing. Just play safe.
			if event.player_index ~= nil then
				-- We shouldn't raise event if the instant-blueprint mod is installed and is activated.
				if not mod_compatibler.is_instant_blueprint_installed_and_activated_for_player_index(event.player_index) then
					if is_entity_ghost then
						util.revive_entity_ghost_and_raise_event(event.created_entity, event.player_index)
					else
						util.revive_tile_ghost_and_raise_event(event.created_entity, event.player_index)
					end
				end
			end
			return
		end
	end
	
	-- Register the built entity if it is one of the creative entities and it needs to be updated via script.
	global_util.register_entity(event.created_entity)
end

-- Callback of the on_tick event.
function events.on_tick()
	-- Auto enable Creative Mode or ask for enabling it if not asked before.
	-- It is not shown in the 0th tick so the message box has to be closed before the popup is shown.
	if not global.creative_mode.enabled then
		if game.tick >= 1 then
			if game.players[1] ~= nil then
				if creative_mode_config.misc.auto_enable then
					global.creative_mode.has_asked_for_enable = true
					cheats.enable_or_disable_creative_mode(true, false, true, game.players[1])
				elseif not global.creative_mode.has_asked_for_enable then
					gui.show_main_popup(game.players[1], gui.main_popup_content_type.enable_creative_mode)
				end
			end
		end
	end
	
	item_source.tick()
	item_void.tick()
	duplicating_chest.tick()
	duplicating_provider_chest.tick()
	void_requester_chest.tick()
	creative_chest_util.tick()
	creative_chest.tick()
	creative_provider_chest.tick()
	fluid_source.tick()
	fluid_void.tick()
	duplicator.tick()
	energy_source.tick()
	passive_energy_source.tick()
	energy_void.tick()
	passive_energy_void.tick()
	
	-- Record the entity each player has opened.
	-- Create or destroy the GUI accordingly.
	for _, player in pairs(game.players) do
		if player.valid then
			local current_opened_entity = player.opened
			local last_opened_entity = global.creative_mode.player_opened_entities[player.index]
			-- Different entities.
			if current_opened_entity ~= last_opened_entity then
				-- Destroy the opened GUI for the last entity.
				gui.create_or_destroy_gui_of_entity(player, last_opened_entity, false)
				-- Create GUI for the currently opened entity if there is any.
				if current_opened_entity ~= nil then
					gui.create_or_destroy_gui_of_entity(player, current_opened_entity, true)
				end
			end
		else
			global.creative_mode.player_opened_entities[player.index] = nil
		end	
	end
	
	-- Instant deconstruction. Iterate from back to front.
	for i = #global.creative_mode.pending_instant_deconstruction, 1, -1 do
		local data = global.creative_mode.pending_instant_deconstruction[i]
		local entity = data.entity
		local player_index = data.player_index
		if entity.valid then
			-- Make sure it is really needed to deconstruct.
			-- Filtered Deconstruction Planner unmarks filtered entities right after they are marked by the deconstruction planner.
			if entity.to_be_deconstructed(entity.force) then
				util.destroy_entity_and_raise_event(entity, player_index)
			end
		end
		table.remove(global.creative_mode.pending_instant_deconstruction, i)
	end
end

-- Returns whether the given entity is a member of the Creative Chest family (i.e. Creative Chest or Creative Provider Chest).
local function is_entity_creative_chest_family(entity)
	-- Creative chest.
	if entity.name == creative_mode_defines.names.entities.creative_chest then
		return true
	end
	-- Creative provider chest.
	if entity.name == creative_mode_defines.names.entities.creative_provider_chest then
		return true
	end
	
	return false
end

-- Returns whether the inventory of the given entity should be cleared before it is mined so that player will not be filled by the items inside it and robots will not need to clear it before deconstruction.
local function should_clear_inventory_before_mined(entity)
	-- Creative chest family.
	if is_entity_creative_chest_family(entity) then
		return true
	end
	
	-- Duplicating chest family.
	if entity.name == creative_mode_defines.names.entities.duplicating_chest then
		return true
	end
	if entity.name == creative_mode_defines.names.entities.duplicating_provider_chest then
		return true
	end
	
	-- Void requester chest.
	if entity.name == creative_mode_defines.names.entities.void_requester_chest then
		return true
	end
	
	return false
end

-- Callback of the on_preplayer_mined_item event, which is invoked just before a player starts mining something.
function events.on_preplayer_mined_item(event)
	-- Clear inventory if needed.
	if should_clear_inventory_before_mined(event.entity) then
		event.entity.get_output_inventory().clear()
	end
end

-- Callback of the on_marked_for_deconstruction event, which is invoked when an entity is marked for deconstruction.
function events.on_marked_for_deconstruction(event)
	-- Clear inventory if needed.
	if should_clear_inventory_before_mined(event.entity) then
		event.entity.get_output_inventory().clear()
	end
	
	-- Thanks Nexela for the inspiration of following snippet.
	-- Instant deconstruction, raise on_entity_died event.
	if global.creative_mode.cheats.instant_deconstruction then
		-- Make sure player_index is provided. Some mods use LuaEntity::order_deconstruction which dosn't not have player_index.
		if event.player_index ~= nil then
			-- But we shouldn't raise event if the instant-blueprint mod is installed and is activated.
			if not mod_compatibler.is_instant_blueprint_installed_and_activated_for_player_index(event.player_index) then
				-- Don't destroy it right now. Filtered Deconstruction Planner may unmark the entity if it is filtered.
				table.insert(global.creative_mode.pending_instant_deconstruction,
				{
					entity = event.entity,
					player_index = event.player_index
				})
			end
		end
	end
end

-- Callback of the on_entity_settings_pasted event, which is invoked after entity copy-paste is done.
function events.on_entity_settings_pasted(event)
	-- If both the source and destination entities are members of the Creative Chest family, move the destination chest to the group of the source chest.
	if is_entity_creative_chest_family(event.source) and is_entity_creative_chest_family(event.destination) then
		creative_chest_util.on_creative_chest_pasted(event.source, event.destination)
	end
end