-- This file contains functions to initialize and update the global meta-table used by this mod.
if not global_util then global_util = {} end

-- Initializes the necessary meta-table if it doesn't exist yet. This is called for new games or existing games when the version of this mod has changed.
function global_util.initialize_or_update_global()
	-- Generate table for storing data that will be used later.
	if not global.creative_mode then
		global.creative_mode = {}
		-- Store whether creative mode has been enabled.
		global.creative_mode.enabled = false
		-- Whether the popup of asking for enabling Creative Mode has been shown.
		global.creative_mode.has_asked_for_enable = false
	else
		if global.creative_mode.enabled then
			global.creative_mode.has_asked_for_enable = true
		end
	end
	-- Also generate the sub-tables to store our entities.
	if not global.creative_mode.item_source_data then global.creative_mode.item_source_data = {} end
	if not global.creative_mode.item_void_data then global.creative_mode.item_void_data = {} end
	if not global.creative_mode.duplicating_chest then global.creative_mode.duplicating_chest = {} end
	if not global.creative_mode.duplicating_chest_next_update_index then global.creative_mode.duplicating_chest_next_update_index = 1 end
	if not global.creative_mode.duplicating_provider_chest then global.creative_mode.duplicating_provider_chest = {} end
	if not global.creative_mode.duplicating_provider_chest_next_update_index then global.creative_mode.duplicating_provider_chest_next_update_index = 1 end
	if not global.creative_mode.void_requester_chest then global.creative_mode.void_requester_chest = {} end
	if not global.creative_mode.void_requester_chest_next_update_index then global.creative_mode.void_requester_chest_next_update_index = 1 end
	if not global.creative_mode.creative_chest_data_groups then global.creative_mode.creative_chest_data_groups = {} end
	if not global.creative_mode.creative_chest_next_group then global.creative_mode.creative_chest_next_group = 1 end
	if not global.creative_mode.creative_provider_chest_data_groups then global.creative_mode.creative_provider_chest_data_groups = {} end
	if not global.creative_mode.creative_provider_chest_next_group then global.creative_mode.creative_provider_chest_next_group = 1 end -- The group number for the next placed creative provider chest. (Number of items in each group <= Chest inventory size)
	if not global.creative_mode.fluid_source then global.creative_mode.fluid_source = {} end
	if not global.creative_mode.fluid_void then global.creative_mode.fluid_void = {} end
	if not global.creative_mode.duplicator_data then global.creative_mode.duplicator_data = {} end
	if not global.creative_mode.energy_source then global.creative_mode.energy_source = {} end
	if not global.creative_mode.passive_energy_source then global.creative_mode.passive_energy_source = {} end
	if not global.creative_mode.energy_void then global.creative_mode.energy_void = {} end
	if not global.creative_mode.passive_energy_void then global.creative_mode.passive_energy_void = {} end
	-- Table for storing which entity each player has opened.
	if not global.creative_mode.player_opened_entities then global.creative_mode.player_opened_entities = {} end
	
	-- Update: item_void -> item_void_data (since v0.1.1)
	if global.creative_mode.item_void then
		for index, item_void in ipairs(global.creative_mode.item_void) do
			-- Re-register the entities again to move them into the new table.
			global_util.register_entity(item_void)
		end
		global.creative_mode.item_void = nil
	end
	
	-- Update: creative_provider_chest_groups -> creative_provider_chest_data_groups (since v0.1.2)
	if global.creative_mode.creative_provider_chest_groups then
		for index, groups in ipairs(global.creative_mode.creative_provider_chest_groups) do
			-- Can't simply re-register in this case because the chests are divided into groups.
			if not global.creative_mode.creative_provider_chest_data_groups[index] then
				global.creative_mode.creative_provider_chest_data_groups[index] = {}
			end
			for index2, chest in ipairs(groups) do
				table.insert(global.creative_mode.creative_provider_chest_data_groups[index], {entity = chest, filtered_slots = {}})
			end
		end
		global.creative_mode.creative_provider_chest_groups = nil
	end
	
	-- Whether Creative Mode has been permanently disabled.
	if not global.creative_mode.permanently_disabled then global.creative_mode.permanently_disabled = false end
	
	-- Table of cheats.
	if not global.creative_mode.cheats then global.creative_mode.cheats = {} end
	-- Instant blueprint.
	if global.creative_mode.cheats.instant_blueprint == nil then
		global.creative_mode.cheats.instant_blueprint = false
		-- instant blueprint is automatically enabled if it is the first time the save has it (Creative Mode < v0.1.2) and also player 1 has cheat mode turned on.
		if global.creative_mode.enabled and game.players[1] ~= nil and game.players[1].cheat_mode then
			cheats.enable_or_disable_instant_blueprint(true)
		end
	end
	-- Instant deconstruction.
	if global.creative_mode.cheats.instant_deconstruction == nil then
		global.creative_mode.cheats.instant_deconstruction = false
		-- same as instant blueprint, instant deconstruction is automatically enabled if...
		if global.creative_mode.enabled and game.players[1] ~= nil and game.players[1].cheat_mode then
			cheats.enable_or_disable_instant_deconstruction(true)
		end
	end
	-- List of entities that are pending to be deconstructed. It is used to encounter Filtered Deconstruction Planner which unmark filtered entities right after they are marked.
	-- A 1-tick delay is used before we really destroy the entities.
	if not global.creative_mode.pending_instant_deconstruction then
		global.creative_mode.pending_instant_deconstruction = {}
	end
	
	-- Also initializes settings.
	global_util.initialize_update_or_keep_config(false)
end

-- Initializes, updates the preserved settings in config.lua, or keeps them as is.
-- @param	force_upate	whether the preserved settings should update their values as the current ones in config.lua
function global_util.initialize_update_or_keep_config(force_update)
	-- Whether we have already saved the config before?
	local has_stored_config_before = global.creative_mode.config ~= nil
	if not has_stored_config_before then
		global.creative_mode.config = {}
	end
	
	-- If we haven't saved the config before or as required, read the newest settings from config.lua.
	local should_read_config = not has_stored_config_before or force_update
	
	-- Preserve settings?
	if not global.creative_mode.config.misc then
		global.creative_mode.config.misc = {}
	end
	if should_read_config or global.creative_mode.config.misc.preserve_settings == nil then
		global.creative_mode.config.misc.preserve_settings = creative_mode_config.misc.preserve_settings
	end
	
	-- If preserve_settings = false, the following settings should be updated for sure.
	should_read_config = should_read_config or global.creative_mode.config.misc.preserve_settings
	
	-- creative_mode_config.misc...
	-- Not used anymore.
	global.creative_mode.config.misc.increase_player_reach_when_enabled = nil
	global.creative_mode.config.misc.increase_player_mining_speed_when_enabled = nil
	global.creative_mode.config.misc.reset_player_reach_when_disabled = nil
	global.creative_mode.config.misc.reset_player_mining_speed_when_disabled = nil
	
	--- creative_mode_config.item_source...
	if not global.creative_mode.config.item_source then
		global.creative_mode.config.item_source = {}
	end
	if should_read_config or global.creative_mode.config.item_source.can_insert_to_vehicle == nil then
		global.creative_mode.config.item_source.can_insert_to_vehicle = creative_mode_config.item_source.can_insert_to_vehicle
	end
	if should_read_config or global.creative_mode.config.item_source.can_insert_to_player == nil then
		global.creative_mode.config.item_source.can_insert_to_player = creative_mode_config.item_source.can_insert_to_player
	end
	if should_read_config or global.creative_mode.config.item_source.insert_only_once_to_player == nil then
		global.creative_mode.config.item_source.insert_only_once_to_player = creative_mode_config.item_source.insert_only_once_to_player
	end
	if should_read_config or global.creative_mode.config.item_source.insert_to_player_amount == nil then
		global.creative_mode.config.item_source.insert_to_player_amount = creative_mode_config.item_source.insert_to_player_amount
	end
	if should_read_config or global.creative_mode.config.item_source.insert_to_player_by_stack == nil then
		global.creative_mode.config.item_source.insert_to_player_by_stack = creative_mode_config.item_source.insert_to_player_by_stack
	end
	if should_read_config or global.creative_mode.config.item_source.can_drop_on_ground == nil then
		global.creative_mode.config.item_source.can_drop_on_ground = creative_mode_config.item_source.can_drop_on_ground
	end
	
	-- creative_mode_config.item_void...
	if not global.creative_mode.config.item_void then
		global.creative_mode.config.item_void = {}
	end
	if should_read_config or global.creative_mode.config.item_void.can_remove_from_vehicle == nil then
		global.creative_mode.config.item_void.can_remove_from_vehicle = creative_mode_config.item_void.can_remove_from_vehicle
	end
	if should_read_config or global.creative_mode.config.item_void.can_remove_from_player == nil then
		global.creative_mode.config.item_void.can_remove_from_player = creative_mode_config.item_void.can_remove_from_player
	end
	if should_read_config or global.creative_mode.config.item_void.can_remove_from_ground == nil then
		global.creative_mode.config.item_void.can_remove_from_ground = creative_mode_config.item_void.can_remove_from_ground
	end
	
	-- creative_mode_config.creative_provider_chest...
	if not global.creative_mode.config.creative_provider_chest then
		global.creative_mode.config.creative_provider_chest = {}
	end
	if should_read_config or global.creative_mode.config.creative_provider_chest.contain_hidden_items == nil then
		global.creative_mode.config.creative_provider_chest.contain_hidden_items = creative_mode_config.creative_provider_chest.contain_hidden_items
	end
	-- Not used anymore.
	global.creative_mode.config.creative_provider_chest.clear_non_used_slots = nil
	
	-- creative_mode_config.duplicator...
	if not global.creative_mode.config.duplicator then
		global.creative_mode.config.duplicator = {}
	end
	if should_read_config or global.creative_mode.config.duplicator.can_duplicate_in_vehicle == nil then
		global.creative_mode.config.duplicator.can_duplicate_in_vehicle = creative_mode_config.duplicator.can_duplicate_in_vehicle
	end
	if should_read_config or global.creative_mode.config.duplicator.can_duplicate_in_player == nil then
		global.creative_mode.config.duplicator.can_duplicate_in_player = creative_mode_config.duplicator.can_duplicate_in_player
	end
end

-- Renews the list of all in-game items and also the data that is used by the Creative Chest family.
function global_util.renew_item_list()
	-- It is not inside the creative_mode table so that it can be reused by other mods... or can it?
	global.item_list = {}
	for _, item in pairs(game.item_prototypes) do
		-- Skip hidden items.
		if global.creative_mode.config.creative_provider_chest.contain_hidden_items or not item.has_flag("hidden") then
			table.insert(global.item_list, item)
		end
	end
	-- Also add back our additional items.
	for _, item_name in ipairs(creative_mode_defines.values.creative_provider_chest_additional_content_names) do
		table.insert(global.item_list, game.item_prototypes[item_name])
	end
	
	-- Update data for the Creative Chest family.
	creative_chest_util.update_item_list_data()
end

-- Initializes the list of all in-game items if it doesn't not exist.
function global_util.initialize_item_list_if_not_exist()
	if not global.item_list then
		global_util.renew_item_list()
	end
end

-- Registers the given entity if needed.
function global_util.register_entity(entity)
	if entity.name == creative_mode_defines.names.entities.item_source then
		table.insert(global.creative_mode.item_source_data,
		{
			entity = entity,									-- The item-source entity.
			slot1_inserted_players = nil,						-- The players who have been inserted the item in the first slot.
			slot1_last_item_position_on_belt = nil,				-- The position of the last inserted item to a transport belt by the first slot.
			slot2_inserted_players = nil,						-- The players who have been inserted the item in the first slot.
			slot2_last_item_position_on_belt = nil,				-- The position of the last inserted item to a transport belt by the first slot.
			last_working_transport_belt = nil,					-- The transport belt it is working for since the last tick. This cache is used for reducing the needs for calling LuaSurface::find_entity.
			last_working_transport_belt_type = 0,				-- The type of the transport belt.
			last_working_static_container = nil,				-- The non-movable container (chest, logistic chest, pipe, assembling machine, rocket silo, furnace, roboport, etc.) it is working for since the last tick.
			last_working_static_container_inventories = nil,	-- The table of inventories that the static container have.
			last_direction = entity.direction					-- The direction of the item source since the last tick. Of course we need this to verify the last working entity.
		})
	elseif entity.name == creative_mode_defines.names.entities.item_void then
		table.insert(global.creative_mode.item_void_data,
		{
			entity = entity,									-- The item-source entity.
			last_working_transport_belt = nil,					-- The transport belt it is working for since the last tick. This cache is used for reducing the needs for calling LuaSurface::find_entity.
			last_working_transport_belt_type = 0,				-- The type of the transport belt.
			last_working_static_container = nil,				-- The non-movable container (chest, logistic chest, pipe, assembling machine, rocket silo, furnace, roboport, etc.) it is working for since the last tick.
			last_working_static_container_inventories = nil,	-- The table of inventories that the static container have.
			last_direction = entity.direction					-- The direction of the item void since the last tick. Of course we need this to verify the last working entity.
		})
	elseif entity.name == creative_mode_defines.names.entities.duplicating_chest then
		table.insert(global.creative_mode.duplicating_chest, entity)
	elseif entity.name == creative_mode_defines.names.entities.duplicating_provider_chest then
		table.insert(global.creative_mode.duplicating_provider_chest, entity)
	elseif entity.name == creative_mode_defines.names.entities.void_requester_chest then
		table.insert(global.creative_mode.void_requester_chest, entity)
	elseif entity.name == creative_mode_defines.names.entities.creative_chest then
		-- Creative Chests are added after Creative Provider Chests with the same logic.
		if not global.creative_mode.creative_chest_data_groups[global.creative_mode.creative_chest_next_place_group] then
			global.creative_mode.creative_chest_data_groups[global.creative_mode.creative_chest_next_place_group] = {}
		end
		table.insert(global.creative_mode.creative_chest_data_groups[global.creative_mode.creative_chest_next_place_group],
		{
			entity = entity,																	-- The creative-chest entity.
			filtered_slots = {},																-- The slot indexes that are filtered out in this chest.
			inventory_display_mode = creative_chest_util.inventory_display_modes.original_mode	-- The inventory display mode.
		})
		-- Increase the group number for the next chest.
		if global.creative_mode.creative_chest_next_place_group < global.creative_mode.creative_provider_chest_num_in_cycle then
			global.creative_mode.creative_chest_next_place_group = global.creative_mode.creative_chest_next_place_group + 1
		else
			global.creative_mode.creative_chest_next_place_group = 1
		end
	elseif entity.name == creative_mode_defines.names.entities.creative_provider_chest then
		-- Creative Provider Chests are divided into groups.
		if not global.creative_mode.creative_provider_chest_data_groups[global.creative_mode.creative_provider_chest_next_place_group] then
			global.creative_mode.creative_provider_chest_data_groups[global.creative_mode.creative_provider_chest_next_place_group] = {}
		end
		table.insert(global.creative_mode.creative_provider_chest_data_groups[global.creative_mode.creative_provider_chest_next_place_group],
		{
			entity = entity,																	-- The creative-chest entity.
			filtered_slots = {},																-- The slot indexes that are filtered out in this chest.
			inventory_display_mode = creative_chest_util.inventory_display_modes.original_mode	-- The inventory display mode.
		})
		-- Increase the group number for the next chest.
		if global.creative_mode.creative_provider_chest_next_place_group < global.creative_mode.creative_provider_chest_num_in_cycle then
			global.creative_mode.creative_provider_chest_next_place_group = global.creative_mode.creative_provider_chest_next_place_group + 1
		else
			global.creative_mode.creative_provider_chest_next_place_group = 1
		end
	elseif entity.name == creative_mode_defines.names.entities.fluid_source then
		table.insert(global.creative_mode.fluid_source, entity)
	elseif entity.name == creative_mode_defines.names.entities.fluid_void then
		table.insert(global.creative_mode.fluid_void, entity)
	elseif entity.name == creative_mode_defines.names.entities.duplicator then
		table.insert(global.creative_mode.duplicator_data,
		{
			entity = entity,									-- The duplicator entity.
			line1_last_item_position_on_belt = nil,				-- The position of the last inserted item on the first line of a transport belt.
			line2_last_item_position_on_belt = nil,				-- The position of the last inserted item on the second line of a transport belt.
			last_working_transport_belt = nil,					-- The transport belt it is working for since the last tick. This cache is used for reducing the needs for calling LuaSurface::find_entity.
			last_working_transport_belt_type = 0,				-- The type of the transport belt.
			last_working_static_container = nil,				-- The non-movable container (chest, logistic chest, assembling machine, rocket silo, furnace, roboport, etc.) it is working for since the last tick.
			last_working_static_container_inventories = nil,	-- The table of inventories that the static container have.
			last_working_static_fluidbox = nil,					-- The non-movable fluidbox (pipe, storage tank, boiler, generator, fluid-turret, etc.) it is working for since the last tick.
			last_direction = entity.direction					-- The direction of the duplicator since the last tick. Of course we need this to verify the last working entity.
		})
	elseif entity.name == creative_mode_defines.names.entities.energy_source then
		table.insert(global.creative_mode.energy_source, entity)
	elseif entity.name == creative_mode_defines.names.entities.passive_energy_source then
		table.insert(global.creative_mode.passive_energy_source, entity)
	elseif entity.name == creative_mode_defines.names.entities.energy_void then
		table.insert(global.creative_mode.energy_void, entity)
	elseif entity.name == creative_mode_defines.names.entities.passive_energy_void then
		table.insert(global.creative_mode.passive_energy_void, entity)
	end
end