-- This file manages remote-interface-related variables or functions that are specific for this mod.
if not remote_interface then remote_interface = {} end

-- Command for showing all available commands for this mod.
local command_help = "/c remote.call(\"" .. creative_mode_defines.names.interface .. "\", \"help\")"
local command_help_list = {}
-- Command for enabling Creative Mode.
local command_enable = "/c remote.call(\"" .. creative_mode_defines.names.interface .. "\", \"enable\")"
table.insert(command_help_list, command_enable)
-- Command for disabling Creative Mode.
local command_disable = "/c remote.call(\"" .. creative_mode_defines.names.interface .. "\", \"disable\", [true|false])"
table.insert(command_help_list, command_disable)
-- Command for updating settings.
local command_update_config = "/c remote.call(\"" .. creative_mode_defines.names.interface .. "\", \"update_config\")"
table.insert(command_help_list, command_update_config)
-- Command for showing the number of creative provider chests per item cycle.
local command_creative_chest_per_cycle = "/c remote.call(\"" .. creative_mode_defines.names.interface .. "\", \"creative_chest_per_cycle\")"
table.insert(command_help_list, command_creative_chest_per_cycle)
-- Command for enabling or disabling the instant blueprint cheat while Creative Mode is enabled.
local command_instant_blueprint = "/c remote.call(\"" .. creative_mode_defines.names.interface .. "\", \"instant_blueprint\", [true|false])"
table.insert(command_help_list, command_instant_blueprint)
-- Command for enabling or disabling the instant deconstruction cheat while Creative Mode is enabled.
local command_instant_deconstruction = "/c remote.call(\"" .. creative_mode_defines.names.interface .. "\", \"instant_deconstruction\", [true|false])"
table.insert(command_help_list, command_instant_deconstruction)


-- The event IDs for the creative mode on-enabled event and on-disabled event respectively.
local on_enabled_event_id
local on_disabled_event_id

-- Returns the event ID for the creative mode on-enabled event.
-- If it has not been loaded, it will be loaded first.
function remote_interface.get_or_load_on_enabled_event_id()
	if on_enabled_event_id == nil then
		on_enabled_event_id = script.generate_event_name()
	end
	return on_enabled_event_id
end
-- Returns the event ID for the creative mode on-disabled event.
-- If it has not been loaded, it will be loaded first.
function remote_interface.get_or_load_on_disabled_event_id()
	if on_disabled_event_id == nil then
		on_disabled_event_id = script.generate_event_name()
	end
	return on_disabled_event_id
end

-- Raises the on_enabled event.
function remote_interface.raise_on_enabled_event()
	game.raise_event(on_enabled_event_id, {})
end

-- Raises the on_disabled event.
function remote_interface.raise_on_disabled_event()
	game.raise_event(on_disabled_event_id, {})
end


-- All remote functions from this mod.
remote_interface.remote_functions =
{
	-- Create a custom event to notify other mods that creative mode is enabled.
	on_enabled = function()
		return remote_interface.get_or_load_on_enabled_event_id()
	end,
	
	-- Create a custom event to notify other mods that creative mode is disabled.
	on_disabled = function()
		return remote_interface.get_or_load_on_disabled_event_id()
	end,
	
	-- The help command.
	help = function()
		game.player.print({"message.creative-mode_available_commands"})
		for _, command in ipairs(command_help_list) do
			game.player.print(command)
		end
	end,
	
	-- The command for enabling creative mode.
	enable = function()
		cheats.enable_or_disable_creative_mode(true, false, false, game.player)
	end,
	
	-- The command for disabling creative mode.
	disable = function(permanent_disable)
		if permanent_disable ~= nil and permanent_disable == true then
			-- If creative mode has not been enabled, notify the player in charge and do nothing.
			if not cheats.check_creative_mode_has_enabled(game.player) then
				return
			end
			gui.show_main_popup(game.player, gui.main_popup_content_type.permanent_disable_creative_mode)
		else
			cheats.enable_or_disable_creative_mode(false, false, false, game.player)
		end
	end,
	
	-- The command for updating the configurable settings.
	update_config = function()
		global_util.initialize_update_or_keep_config(true)
		local player_name
		if game.player.name == nil then
			player_name = {"entity-name.player"}
		else
			player_name = game.player.name
		end
		util.print_to_all_players({"message.creative-mode_config-updated", player_name})
	end,
	
	-- The command for printing the number of creative provider chests per item cycle.
	creative_chest_per_cycle = function()
		game.player.print({"message.creative-mode_num-creative-provider-chest-per-item-cycle", global.creative_mode.creative_provider_chest_num_in_cycle})
	end,
	
	-- The command for enabling or disabling the instant blueprint cheat while Creative Mode is enabled.
	instant_blueprint = function(enable)
		if enable == nil then
			enable = true
		end
		-- If player is trying to enable the cheat but creative mode has not been enabled, notify him and do nothing.
		if enable and not cheats.check_creative_mode_has_enabled(game.player) then
			return
		end
		cheats.enable_or_disable_instant_blueprint(enable)
	end,
	
	-- The command for enabling or disabling the instant deconstruction cheat while Creative Mode is enabled.
	instant_deconstruction = function(enable)
		if enable == nil then
			enable = true
		end
		-- If player is trying to enable the cheat but creative mode has not been enabled, notify him and do nothing.
		if enable and not cheats.check_creative_mode_has_enabled(game.player) then
			return
		end
		cheats.enable_or_disable_instant_deconstruction(enable)
	end
}