-- This file contains variables or functions that are not specific to any mod and can be used by other mods.
if not util then util = {} end

-- Cache the reference to math functions can speed up the process slightly.
-- When called 1216 times in each frame, this can boost the FPS from 57 to 58.
local math_floor = math.floor
local math_ceil = math.ceil
local math_max = math.max
local math_min = math.min
local math_fmod = math.fmod

-- The distance between items on transport belt.
-- Last update: Factorio 0.13.3
transport_belt_item_distance = 0.28125

-- Limits the given number in between the given minimum and maximum inclusively.
function util.clamp(number, minimum, maximum)
	return math_max(minimum, math_min(number, maximum))
end

-- Wraps the given number between the given minimum and maximum inclusively.
function util.repeat_index(number, minimum, maximum)
	local range_size = maximum - minimum + 1
	if number < minimum then
		number = number + (range_size * (math_floor((minimum - number) / range_size) + 1))
	end
	return minimum + math_fmod(number - minimum, range_size)
end

-- Returns whether the given sorted table contain the given specific value.
function util.has_value_sorted(tab, val)
	for _, v in ipairs(tab) do
		if v == val then
			return true
		end
	end
	return false
end

-- Prints the given message to all players.
function util.print_to_all_players(message)
	for _, player in pairs(game.players) do
		player.print(message)
	end
end

-- Returns the bounding box of the tile at the given position.
-- Note: turns out it is quite a FPS heavy task. Use only when needed.
-- @param	position	It should contain {x: float, y: float}
function util.get_tile_bb(position)
	return
	{
		{math_floor(position.x), math_floor(position.y)},
		{math_ceil(position.x), math_ceil(position.y)}
	}
end

-- Returns whether the given entity has at least one inventory.
function util.has_inventory(entity)
	for _, inv in pairs(defines.inventory) do
		if entity.get_inventory(inv) ~= nil then
			return true
		end
	end
	return false
end

-- Returns whether the given entity has at least one fluidbox.
function util.has_fluidbox(entity)
	if not entity.fluidbox then
		return false
	end
	return #entity.fluidbox > 0
end

-- Returns whether the given inserter entity is enabled according to its circuit network state as well as its logistic network state.
function util.is_inserter_enabled(inserter)
	local control = inserter.get_control_behavior()
	-- Does it have control behaviour? (Not connected = no control?)
	if control and control.valid then
		-- Check logistic network.
		-- Is logistic network connected?
		if control.connect_to_logistic_network then
			-- Has logistic condition set?
			if control.logistic_condition then
				-- Condition fulfilled?
				if not control.logistic_condition.fulfilled then
					return false
				end
			else
				-- Connected but no condition, hence not OK.
				return false
			end
		end
	
		-- Check circuit network.
		-- Is connected by wire?
		if control.get_circuit_network(defines.wire_type.red, defines.circuit_connector_id.inserter) or control.get_circuit_network(defines.wire_type.green, defines.circuit_connector_id.inserter) then
			-- Is operation mode enable/disable?
			if control.circuit_mode_of_operation == defines.control_behavior.inserter.circuit_mode_of_operation.enable_disable then
				-- Has the circuit condition set?
				if control.circuit_condition then
					-- Condition fulfilled?
					if not control.circuit_condition.fulfilled then
						return false
					end
				else
					-- Connected but no condition, hence not OK.
					return false
				end
			end
		end
	end
	
	-- No control? Because not connected to network?
	return true
end

-- Revives the given entity ghost and raises the on_built_entity event for it.
-- You shouldn't call this function if the given ghost is actually a tile ghost!
function util.revive_entity_ghost_and_raise_event(entity_ghost, reviver_player_index)
	-- Get the items this ghost will request when it is revived, most likely modules.
	local item_requests = entity_ghost.item_requests
	-- Nullify the requests because we will supply them.
	entity_ghost.item_requests = nil
	-- Revive entity.
	local _, revived_entity = entity_ghost.revive()
	if revived_entity and revived_entity.valid then
		-- Supply the requested items.
		if item_requests then
			-- Try module inventory slots first.
			local module_inventory = revived_entity.get_module_inventory()
			if module_inventory then
				for index, item in pairs(item_requests) do
					-- Reduce the number of items being inserted.
					item.count = item.count - module_inventory.insert{name = item.item, count = item.count}
				end
			end
			-- In case there are items other than modules left (special items?), here is the last chance to insert them to the entity.
			for index, item in pairs(item_requests) do
				if item.count > 0 then
					revived_entity.insert{name = item.item, count = item.count}
				end
			end
		end
		
		-- Raise event.
		game.raise_event(defines.events.on_built_entity,
		{
			player_index = reviver_player_index,
			created_entity = revived_entity
		})
	end
end

-- Revives the given tile ghost and raises the on_built_entity event for it.
function util.revive_tile_ghost_and_raise_event(tile_ghost, reviver_player_index)
	-- A simplified version of the entity ghost revive function, as tile doesn't have item requests.
	local _, revived_entity = tile_ghost.revive()
	if revived_entity and revived_entity.valid then
		game.raise_event(defines.events.on_built_entity,
		{
			player_index = reviver_player_index,
			created_entity = revived_entity
		})
	end
end

-- Destroys the given entity and raises the on_entity_died event for it.
function util.destroy_entity_and_raise_event(entity, destroyer_player_index)
	game.raise_event(defines.events.on_entity_died,
	{
		entity = entity,
		force = game.players[destroyer_player_index].force
	})
	entity.destroy()
end